import numpy as np

def intersect(x=[], y=[], intersect_value=1., deg=1):
    """
    Calculates de intersection between two binned distributions.
    Returns 
    1. The abcissa of the input data point closest to the intersection.
    2. A boolean that says whether there was at least one intersection or not.
    """
    if len(x)<2 or len(y)<2:
        raise ValueError('Please provide the x and y data.')
    _polyfit = np.polyfit(x, y, deg=deg)
    _func = np.poly1d(_polyfit)
    _x2 = np.linspace(x[0], x[-1], 10000)
    _y2 = _func(_x2)
    _idx = np.argwhere(np.diff(np.sign(_y2 - intersect_value))).flatten()
    if len(_idx)==0:
        return -1, False
    _idx = int(_idx[-1]) #consider the last intersection index only
    if abs(_y2[_idx]-intersect_value) < abs(_y2[_idx+1]-intersect_value):
        return _x2[_idx], True
    else:
        return _x2[_idx+1], True

def is_list_empty(l):
    if isinstance(l, list): #it is a list
        return all( map(is_list_empty, l)) #all([]) is True
    return False



#MISSING: Add methods to retrieve the halos and subhalos stored in backup

import os, warnings
import numpy as np
import pynbody as pn
import pynbody.units as u
from pynbody.analysis.halo import center_of_mass as com
from dmprofile.fit import nfw_fit
from dmprofile.utilities import intersect
from dmprofile.move import centering_com

class Halos():
    """
    Halos management
    """
    def __init__(self, filename, N=1):
        self.filename = filename
        _s = pn.load(os.path.join(self.filename))
        _s['eps'] = 200.*pn.units.pc
        _s.physical_units() #concentration will be wrong if this line is removed
        if N>len(_s): 
            warnings.warn("The specified halo number is larger than the number of halos in the simulation box.")
        self._N = N
        self._halos = _s.halos()[:self._N]
        self._halos_bak = self._halos #halos backup

        #backup for subhalos (id, subhalo): 
        #the id is helpful because each halo can have more than one subhalo
        #'0' is the id for the main subhalo, '1' is for the second largest, etc
        self._subhalos = [[] for _ in range(self._N)] #one row per halo in descending size order
        for _i in range(self._N):
            length = int(self._halos[_i].properties['NsubPerHalo'])#number subhalos in a halo
            #columns represent the subhalos in descending size order
            self._subhalos[_i].extend((_isub, self._halos[_i].sub[_isub]) for _isub in range(length))
        self._subhalos_bak = self._subhalos
    
    def _get_r200(self, idx):
        return self._halos[idx].properties['Halo_R_Crit200'].in_units('kpc')

    def _check_backup(self, idx, sub_idx=-1):
        if sub_idx < 0:
            return self._halos[idx]==self._halos_bak[idx]
        else:
            return self._subhalos[idx][sub_idx]==self._subhalos_bak[idx][sub_idx]

    def restore(self, idx, sub_idx=-1):
        if idx >= self._N: raise ValueError('The halo index is too large.')
        if type(idx) != int or type(sub_idx) != int: raise TypeError('The indexes must be integers.')
        if sub_idx<0:
            self._halos[idx], self._halos_bak[idx] = self._halos_bak[idx], self._halos[idx]
        else:
            self._subhalos[idx][sub_idx], self._subhalos_bak[idx][sub_idx] = self._subhalos_bak[idx][sub_idx], self._subhalos[idx][sub_idx]

    def filter_(self, halo_idxs, sub_idx=-1, filter_str='Sphere_1.2', option=None):
        """
        halo_idxs: indexes of the halos to filter; if more than one, it must be a list
        option: None, 'half1', 'half2' or 'all'
        filters:
        Sphere_radius, where radius is a number relative to r200. Example: Sphere_1
        BandPass_prop_min_max. Example: BandPass_y_1 kpc_2 kpc. It must have units.
        The inputs for each filter are separated using the underscore '_' carachter
        More filters can be added to 'filter_dict' in the same fashion with variable number of arguments
        """
        if type(halo_idxs) != int and type(halo_idxs) != list:
            raise TypeError('The introduced indexes have the wrong format')
        if type(halo_idxs) is int:
            halo_idxs = [halo_idxs]

        filter_dict = {'Sphere': lambda _r,_factors,_halo: 
                       pn.filt.Sphere(float(_factors[0])*_r, com(_halo)),
                       'BandPass': lambda _r,_factors,_halo: 
                       pn.filt.BandPass(_factors[0], _factors[1], _factors[2])}
        for item in filter_dict.keys():
            if item in filter_str:
                break
            else:
                if item==[*filter_dict.keys()][-1]:
                    raise ValueError('The specified filtering option is not supported.')
        _split = filter_str.split('_')
        _split_str = _split[0]
        _split_values = {i: s for i,s in enumerate(_split[1:])}
        print("---FILTER--- Type:", _split_str, "-- Parameters:", *_split_values.values())
        _filter_func = filter_dict[_split_str]

        if option=='all': halo_idxs=range(self._N) 
        elif option=='half1': halo_idxs=range(int(self._N/2))
        elif option=='half2': halo_idxs=range(int(self._N/2),self._N)
        elif option==None: pass
        else: raise ValueError('That option is not supported')

        for i in halo_idxs:
            _r200 = self._get_r200(i)
            
            if sub_idx < 0: #filter the halo
                if not self._check_backup(i):
                    print(self._check_backup(i))
                    raise warnings.warn('This halo backup is already being used!')
                self._halos_bak[i] = self._halos[i]
                with centering_com(self._halos[i]):
                    self._halos[i] = self._halos[i][_filter_func(_r200, _split_values, 
                                                                 self._halos[i])]
            else: #filter the halo but centered in the subhalo
                if not self._check_backup(i):
                    raise warnings.warn('This subhalo backup is already being used!')
                if self._subhalos[i][sub_idx][0] != sub_idx:
                    raise RuntimeError('Different subhalos are being mixed!')
                self._subhalos_bak[i][sub_idx] = self._subhalos[i][sub_idx]
                with centering_com(self._subhalos[i][sub_idx][1]):
                    _subhalo = self._halos[i][_filter_func(_r200, _split_values, 
                                                           self._subhalos[i][sub_idx][1])]
                    self._subhalos[i][sub_idx] = (sub_idx,_subhalo)


    def is_relaxed(self, halo_idx, sub_idx=-1):
        if sub_idx < 0:
            if len(self._subhalos[halo_idx])==0:
                raise RuntimeError('This halo has no subhalos!')
            _h = self._halos[halo_idx]
        else:
            _h = self._halos[halo_idx].sub[sub_idx]
        with centering_com(self._halos[halo_idx]): #centering the main halo
            _com = pn.analysis.halo.center_of_mass(_h)
            _r200 = self._get_r200(halo_idx)
            _dist = np.sqrt( np.power(_h[0]['pos']-_com,2).sum() )
            #The minimum potential point is obtained with:
        #halo[halo_index].subhalo[subhalo_index][0]['pos']
        return bool(_dist < 0.07*_r200)

    def is_resolved(self, halo_idx, intersect_value=1.):
        with centering_com(self._halos[halo_idx]): #centering the main halo
            _r200 = self._get_r200(halo_idx)
            _prof = self.get_profile(halo_idx, 'dm', bins=(2.4,25,30), bin_type='log', normalize=False)
            _relax = self.relaxation(halo_idx, _prof)
            _inter = intersect(_prof['rbins'], _relax, intersect_value=intersect_value)
            if _inter[1]==True:            
                print(_inter[0], " ---- ", _r200*np.power(10,-1.25))
                return _inter[0] < _r200*np.power(10,-1.25)
            else:
                raise RuntimeError('No intersection was found for this halo. The resolution criteria cannot thus be checked.')
                return False

    def get_halo(self, halo_idx):
        if halo_idx<=self._N: 
            return self._halos[halo_idx]
        else:
            raise IndexError("The halo you requested was not stored.")
            return self._halos[-1]

    def get_subhalo(self, halo_idx, sub_idx=0):
        """
        By default returns the largest subhalo in the specified halo.
        """
        return self._subhalos[halo_idx][sub_idx][1]

    def get_subhalos(self, sub_idx=-1):
        """
        Either returns all subhalos or just the subhalos with a specific index for all halos.
        The subhalos will be wrapped in a tuple with their respective indexes.
        """
        if sub_idx<0:
            return self._subhalos
        else:
            return [self._subhalos[i][sub_idx] for i in range(self._N)]

    def get_halos(self):
        return self._halos

    def concentration_200(self, idx, sub_idx=-1):
        """
        Calculates the concentration at r200 
        """
        if sub_idx<0:
            _current_halo = self._halos[idx] 
        else:
            _current_halo = self._subhalos[idx][sub_idx][1]

        with centering_com(self._halos[idx]):
            #we always use the main halo for the r200 distance, even for subhalos
            _r200 = _get_r200(idx)
            _profile = pn.analysis.profile.Profile(_current_halo.dm, 
                                                   ndim=3, nbins=250, min=0.5, max=35)
            _rs = nfw_fit(_profile)[1]
            print(_r200, _rs)
        return _r200/_rs
      
    def concentrations_200(self, sub='False', sub_idx=0, n=None):
        """
        Calculates the first 'n' concentrations at r200. Returns a list.
        """
        if n==None: 
            n = self._N
        return [self.concentration_200(i, sub, sub_idx) for i in range(n)]

    def get_profile(self, idx, sub_idx=-1, component='dm', 
                    bins=(1,45,30), bin_type='linear', normalize=False):
        """
        Obtain the profile of a single halo.
        Arguments:
        'idx': which halo to get the profile from
        'sub_idx': which subhalo to get the profile from
        'component': 'dm', 'stars', 'gas' or 'all'
        'bins': tuple with (start,stop,nbins)
        'bin_type': either 'linear' or 'log'
        'normalize': 'None' for no normalization and 'r200' for normalization
        'centering': whether to centre the halo before obtaining the profile or not
        """
        _h = self._halos[idx] if sub_idx<0 else self._subhalos[idx][sub_idx][1]
        with centering_com(_h):
            components = {'dm':_h.dm, 'stars':_h.s, 'gas':_h.g, 'all':_h}
            if component not in components:
                raise ValueError('The specified component does not exist.')
            if type(bins) is not tuple or len(bins) != 3:
                raise TypeError('Please insert bins with the following pattern: (start, stop, nbins).')
        
            if normalize:
                r200 = self._get_r200(idx)
                calc_x = lambda x: x['r']/r200
            else:
                calc_x = lambda x: x['r']
            
            if bin_type=='linear':
                return pn.analysis.profile.Profile(components[component],
                                                   calc_x = calc_x,
                                                   ndim=3, 
                                                   bins=np.linspace(bins[0],
                                                                    bins[1],
                                                                    bins[2]))
            elif bin_type=='log':
                return pn.analysis.profile.Profile(components[component], 
                                                   calc_x = calc_x,
                                                   ndim=3, 
                                                   bins=np.logspace(np.log10(bins[0]),
                                                                    np.log10(bins[1]),
                                                                    bins[2]))
            else:
                raise ValueError('The specified bin type does not exist.')

    def relaxation(self, idx, profile=None, component='dm', bins=(4,50,30), bin_type='linear'):
        """
        Creates a t_relax(r) / t_circ(r200) array for the provided halo.
        If you do not provide a profile, the method creates one for you. 
        The profile is needed to obtain the density as a function of r.
        """
        _h = self._halos[idx]
        _rho_crit_z = pn.array.SimArray(pn.analysis.cosmology.rho_crit(_h, unit="Msol kpc**-3"), 
                                        "Msol kpc**-3")
        with centering_com(_h):
            if profile==None:
                _prof = self.get_profile(idx, component=component, bins=bins, 
                                         bin_type=bin_type, normalize=False)
                _density = _prof['density']
                _rbins = _prof['rbins']
                print(_density)
                print(_rbins)
            else:
                _density = profile['density']
                _rbins= profile['rbins']

            _nparts = pn.array.SimArray([len(_h[_h['r']<_rbins[i]]) for i in range(len(_rbins))],'')

        if 0. in _density:
            print(_density)
            raise ValueError('At least one of the values of the density is zero. The relaxation time cannot be calculated when this happens.')
        return (np.sqrt(200)*_nparts)/(8*np.log(_nparts))*np.sqrt(_rho_crit_z/_density)



import matplotlib
matplotlib.use('Agg')
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import itertools, collections

from pynbody.analysis.theoretical_profiles import NFWprofile
from dmprofile.fit import nfw_fit
from dmprofile.utilities import intersect, is_list_empty

#decorator: applies the same function to all objects in _Plot._obj_list
def do_all_objects(func):
    def wrapper(self, *args, **kwargs):
        kwargs['i_profile'] = 0
        for i in range(self._nrows):
            for j in range(self._ncols):
                if kwargs['i_profile'] < self._N:
                    kwargs['i'] = i
                    kwargs['j'] = j
                    func(self, *args, **kwargs)
                    kwargs['i_profile'] += 1
    return wrapper


class _Plot():
    """
    Important: each individual object contained inside obj_list is meant to be converted into a plot, so in general each object will be a list of values.
    """
    def __init__(self, obj_list, width, height, nrows, ncols, name=''):
        if type(obj_list) is not list:
            raise TypeError
        if type(width) is not int or type(height) is not int or type(nrows) is not int or type(ncols) is not int:
            raise TypeError
        self._obj_list = obj_list
        self._N = len(self._obj_list)
        self._nrows, self._ncols = nrows, ncols
        self._name = name

    def _geometry(self, n):
        """
        Returns number of rows '_r' and columns '_c' that any plot with more than 2 figure has.
        'n' is the number of figures to plot
        """
        _r = np.floor(np.sqrt(n))
        _c = _r
        _r_iter, _c_iter = itertools.cycle(range(2)), itertools.cycle(range(2))
        next(_r_iter)
        while _r*_c<self._N:
            _r += next(_r_iter)
            _c += next(_c_iter)
        return int(_r), int(_c)
        
    def _set_axis_indexes(self, axis_idx):
        """
        Set the axis indexes into 1D or 2D, due to matplotlib axis indexing.
        1D is needed when one plans to plot less than two pictures.
        """
        if self._N>2:
            return axis_idx[0], axis_idx[1]
        else:
            return axis_idx[0]

    def set_title(self, title):
        self._fig.suptitle(title)
        
    def set_name(self, name):
        self._name = name

    def set_axis(self, idx, xlabel, ylabel, xscale='linear', yscale='linear'):
        """
        idx can be either a single value or a tuple, depending on the total number of profiles
        """
        if self._N>2:
            indexes = idx[0], idx[1]
        else:
            indexes = idx[0]
        self._axis[indexes].set_xscale(xscale)
        self._axis[indexes].set_yscale(yscale)
        self._axis[indexes].set_xlabel(xlabel)
        self._axis[indexes].set_ylabel(ylabel)

    def set_axis_all(self, xlabel, ylabel, xscale='linear', yscale='linear'):
        if self._N>2:
            _r, _c = self._geometry(self._N)
            for i in range(_r):
                for j in range(_c):
                    self.set_axis((i,j), xlabel, ylabel, xscale, yscale)
        else:
            for i in range(self._N):
                self.set_axis((i,0), xlabel, ylabel, xscale, yscale)

    def set_all_properties(self, title="", model=None, 
                       xlabel='R [kpc]', ylabel=None, 
                       xscale='linear', yscale='linear'):
        """
        model options: 'density_profile', 'mass_enc_profile', 'mass_profile',
                       'density_enc_profile', 'relaxation', 'concentration_mass'
        """
        self.set_title(title)
        scale_options = ['linear', 'log']
        if xscale not in scale_options or yscale not in scale_options:
            raise ValueError('The selected scale is not supported.')
        model_options = ['density_profile', 
                         'density_enc_profile',
                         'mass_enc_profile', 
                         'mass_profile',
                         'concentration_mass',
                         'relaxation']
        if model not in model_options:
            raise ValueError('The specified model is not currently predefined.')
        if model!='None':
            self.set_axis_all('R [kpc]', r'$\rho$ [M$_{\odot}$ kpc$^{-3}$]', xscale, yscale) if model=='density_profile' else self.set_axis_all('R [kpc]', r'$\rho_{enc}$ [M$_{\odot}$ kpc$^{-3}$]', xscale, yscale) if model=='density_enc_profile' else self.set_axis_all('R [kpc]', r'M$_{enc}$ [M$_{\odot}$]', xscale, yscale) if model=='mass_enc_profile' else self.set_axis_all(r'M$_{200}$ [M$_{\odot}$]', r'c$_{200}$', xscale, yscale) if model=='concentration_mass' else self.set_axis_all('R [kpc]', r't$_{relax}(R)$/t$_{circ}$(r$_{200}$)', xscale, yscale) if model=='relaxation' else self.set_axis_all('R [kpc]', r'M [M$_{\odot}$]', xscale, yscale)
        else:
            for i in range(self._nrows):
                if self._N>2:
                    for j in range(self._ncols):
                        self.set_axis((i,j), xlabel, ylabel, xscale, yscale)
                else:
                    self.set_axis((i,0), xlabel, ylabel, xscale, yscale)
                    
    def savefig(self, name=''):
        if self._name=='':
            try:
                self._fig.savefig(name)
            except ValueError:
                print("A figure cannot be saved without having a name.")
        else:
            self._fig.savefig(self._name)


class Profile(_Plot):
    """
    Plots all kinds of pynbody related profiles.

    Args:
    p: list of pynbody profile objects
    """
    def __init__(self, p, w=10, h=9, nrows=0, ncols=0, name=''):
        super().__init__(p, w, h, nrows, ncols, name)
        self._p = self._obj_list
        if nrows==0 or ncols==0:
            self._nrows, self._ncols = super()._geometry(self._N)
        self._fig, self._axis = plt.subplots(nrows=self._nrows, ncols=self._ncols, figsize=(w,h))

    def plot(self, idx, axis_idx, x_var, y_var):
        """
        Args:
        1. 'idx':  profile index (int)
        2. 'axis_idx': axis index (int [self._N<=2] or tuple [self._N>2])
        3. 'x_var': variable stored in the profile to plot as x variable
            options: radius, mass, mass_enc, density
        4. 'y_var': variable stored in the profile to plot as y variable
            options: radius, mass, mass_enc, density
        """
        def _define_vars(x_var, y_var):
            _translate = {'radius': self._p[idx]['rbins'],
                          'mass': self._p[idx]['mass'],
                          'mass_enc': self._p[idx]['mass_enc'],
                          'density': self._p[idx]['density'],
                          'density_enc': (3*self._p[idx]['mass_enc'])
                          /(4*np.pi*np.power(self._p[idx]['rbins'],3))}
            if x_var not in _translate.keys() or y_var not in _translate.keys():
                raise ValueError
            return _translate[x_var], _translate[y_var]

        _x, _y = _define_vars(x_var, y_var)
        indexes = super()._set_axis_indexes(axis_idx)
        self._axis[indexes].plot(_x, _y)

    @do_all_objects
    def plot_all(self, x_var, y_var, *args, **kwargs):
        self.plot(kwargs['i_profile'], (kwargs['i'], kwargs['j']), x_var, y_var)

    def fit_and_plot(self, idx, axis_idx, function='nfw'):
        functions_implemented = ['nfw']
        if function not in functions_implemented:
            raise ValueError('The specified function has not yet been implemented.')
        
        if function=='nfw':
            _fit = nfw_fit(self._p[idx])
            _func = NFWprofile.profile_functional_static(np.array(self._p[idx]['rbins']), 
                                                         _fit[0], _fit[1])
            indexes = super()._set_axis_indexes(axis_idx)
            self._axis[indexes].plot(np.array(self._p[idx]['rbins']), 
                                     _func, label='NFW fit')
            self._axis[indexes].legend()

    @do_all_objects
    def fit_and_plot_all(self, function='nfw', *args, **kwargs):
        self.fit_and_plot(kwargs['i_profile'], (kwargs['i'], kwargs['j']))

        
    def draw_line(self, axis_idx, value, orientation='v', label='---', color='grey', 
                  linestyle = '--'):
        """
        Arguments:
        'idx': profile index
        'axis_idx' index of the axis where the line will be plotted
        'value': value at which will be plotted; it is related to the 'orientation' option ('v' or 'h')
        """
        indexes = super()._set_axis_indexes(axis_idx)
        _xmin, _xmax = self._axis[indexes].get_xbound()
        _ymin, _ymax = self._axis[indexes].get_ybound()
        if orientation=='v':
            self._axis[indexes].plot([value,value],[_ymin,_ymax], 
                                     linestyle=linestyle, color=color, label=label)
        elif orientation=='h':
            self._axis[indexes].plot([_xmin,_xmax],[value,value],
                                     linestyle=linestyle, color=color, label=label)
        else:
            raise ValueError('The specified orientation is not supported.')
        self._axis[indexes].legend()

class Shape(_Plot):
    """
    Plots all kinds of pynbody related shapes.

    Args:
    s: list of shape arrays for plotting, i.e., list of lists of pynbody shape objects
    extra_var: list of some additional property common to all objects from which the shapes were taken
           for example: radius, mass
    """
    def __init__(self, s, extra_var=[], w=10, h=9, nrows=0, ncols=0, name=''):
        super().__init__(s, w, h, nrows, ncols, name)
        self._s = self._obj_list
        self._extra_var = extra_var
        if nrows==0 or ncols==0:
            self._nrows, self._ncols = super()._geometry(self._N)
        self._fig, self._axis = plt.subplots(nrows=self._nrows, ncols=self._ncols, figsize=(w,h))

    def _shape_array_to_numpy(self, s):
        """
        Converts the shape formatting to allow correct indexing.
        This is needed for plotting.
        """
        _snew = []
        for i in range(len(s)):
            _a, _b, _c, _d, _e, _f = ([] for i in range(6))
            for j in range(len(s[i])):
                _a.append(s[i][j][0][0]) #pos
                _b.append(s[i][j][1][0]) #b/a
                _c.append(s[i][j][2][0]) #c/a
                _d.append(s[i][j][3][0]) #alignment
                _e.append(s[i][j][4][0]) #rotation matrix
                _f.append((1-s[i][j][1][0]**2)/(1-s[i][j][2][0]**2)) if 1-s[i][j][2][0]**2<1 else _f.append(1.1) #triaxiality 
            _snew.append([_a,_b,_c,_d,_e,_f])
        return _snew

    def scatter_plot(self, idx, axis_idx, x_var, y_var, resolved_bools=[], relaxed_bools=[]):
        """
        Args:
        1. 'idx': pynbody shape list object. It refers to the way self._s is ordered
        2. 'axis_idx': axis index (int [self._N<=2] or tuple [self._N>2])
        3. 'x_var': variable stored in the shape to plot as x variable
            options: pos, b/a, c/a, align, rot, triax, m200, r200
        4. 'y_var': variable stored in the profile to plot as y variable
            options: pos, b/a, c/a, align, rot, triax, m200, r200
        """
        handles = []
        indexes = super()._set_axis_indexes(axis_idx)

        def _define_vars(x_var, y_var):
            """
            Defines which arrays are to be plotted according to user input.
            See online documentation of pynbody.analysis.halo.halo_shape() method.
            """
            _s_for_plot = self._shape_array_to_numpy(self._s)
            _match = {'position': 0,
                      'b/a': 1,
                      'c/a': 2,
                      'align': 3,
                      'rot': 4,
                      'triax': 5}
            if x_var not in _match.keys() and y_var not in _match.keys():
                raise ValueError("At least one of the variables being plot must be related to the shape pynbody method.")
            elif x_var in _match.keys() and y_var in _match.keys():
                return _s_for_plot[idx][_match[x_var]], _s_for_plot[idx][_match[y_var]]
            else: #either x or y are not contained inside the self._s object
                if x_var not in _match.keys():
                    return self._extra_var, _s_for_plot[idx][_match[y_var]]
                else:
                    return _s_for_plot[idx][_match[x_var]], self._extra_var

        if not is_list_empty(resolved_bools):
            grey = (0.2,0.2,0.2)
            lightgrey = (0.6,0.6,0.6)
            resolution_colors = [lightgrey if resolved_bools[i]==False else grey for i in range(len(self._s[idx]))]
            handles.append(mpatches.Patch(color=lightgrey, label='unresolved structures'))
            handles.append(mpatches.Patch(color=grey, label='resolved structures'))
        else:
            resolution_colors = [(0.4,0.4,0.4)]*len(self._s[idx])

        if not is_list_empty(relaxed_bools):
            relaxation_markers = ['o' if relaxed_bools[i]==False else 'v' for i in range(len(self._s[idx]))]  
            handles.append(plt.plot([],[], marker='o',ms=8,c='k',
                                    linestyle='None',label='unrelaxed structure')[0])
            handles.append(plt.plot([],[], marker='v',ms=8,c='k',
                                    linestyle='None',label='relaxed structure')[0])
            
            self._axis[indexes].legend(handles=handles)
        else:
            relaxation_markers = ['v']*len(self._s[idx])
        
        _x, _y = _define_vars(x_var, y_var)
        _sorted_lists = sorted(zip(_x,_y), key=lambda x: x[0])
        _x, _y = [[q[i] for q in _sorted_lists] for i in range(2)]
        
        for xp, yp, c, m in zip(_x, _y, resolution_colors, relaxation_markers):
            self._axis[indexes].scatter(xp, yp, c=c, marker=m)


class Concentration(_Plot):
    """
    Plots all kinds of pynbody related concentrations.

    Arguments:
    c: list of lists of concentration values
    extra_var: list of some additional property common to all objects from which the shapes were taken
           for example: radius, mass
    """
    def __init__(self, c, extra_var=[], w=10, h=9, nrows=0, ncols=0, name=''):
        super().__init__(c, w, h, nrows, ncols, name)
        self._c = self._obj_list
        self._extra_var = extra_var
        if nrows==0 or ncols==0:
            self._nrows, self._ncols = super()._geometry(self._N)
        self._fig, self._axis = plt.subplots(nrows=self._nrows, ncols=self._ncols, figsize=(w,h))

    def scatter_plot(self, idx, axis_idx, concentration_axis='y', 
                     resolved_bools=[], relaxed_bools=[]):
        """
        Args:
        'resolved_bools': Resolved halos. The order refers to the concentration array included when calling the class constructor.
        'relaxed_bools': Relaxed halos. The order refers to the concentration array included when calling the class constructor.
        """
        axis_options = ['x','y']
        handles = []
        if concentration_axis not in axis_options:
            raise ValueError('The axis you selected for plotting the concentration is not valid.')
        if idx >= self._N:
            raise ValueError('You cannot plot something that does not exist :)')
        indexes = super()._set_axis_indexes(axis_idx)
        
        if not is_list_empty(resolved_bools):
            grey = (0.2,0.2,0.2)
            lightgrey = (0.6,0.6,0.6)
            resolution_colors = [lightgrey if resolved_bools[i]==False else grey for i in range(len(self._c[idx]))]
            handles.append(mpatches.Patch(color=lightgrey, label='unresolved structures'))
            handles.append(mpatches.Patch(color=grey, label='resolved structures'))
        else:
            resolution_colors = [(0.4,0.4,0.4)]*len(self._c[idx])

        if not is_list_empty(relaxed_bools):
            relaxation_markers = ['o' if relaxed_bools[i]==False else 'v' for i in range(len(self._c[idx]))]  
            handles.append(plt.plot([],[], marker='o',ms=8,c='k',
                                    linestyle='None',label='unrelaxed structure')[0])
            handles.append(plt.plot([],[], marker='v',ms=8,c='k',
                                    linestyle='None',label='relaxed structure')[0])
            
            self._axis[indexes].legend(handles=handles)
        else:
            relaxation_markers = ['v']*len(self._c[idx])
        for xp, yp, c, m in zip(self._c[idx], self._extra_var, resolution_colors, relaxation_markers):
            if concentration_axis=='x':
                self._axis[indexes].scatter(xp, yp, c=c, marker=m)
            else:
                self._axis[indexes].scatter(yp, xp, c=c, marker=m)

class Relaxation(_Plot):
    """
    Args:
    relax: list of pynbody relaxation arrays (obtained for example from dmprofile.halos.relaxation)
    r: corresponding list of radius
    """
    def __init__(self, relax, radius, w=10, h=9, nrows=0, ncols=0, name=''):
        super().__init__(relax, w, h, nrows, ncols, name)
        self._relax = self._obj_list
        self._radius = radius
        if nrows==0 or ncols==0:
            self._nrows, self._ncols = super()._geometry(self._N)
        self._fig, self._axis = plt.subplots(nrows=self._nrows, ncols=self._ncols, figsize=(w,h))

    def plot(self, idx, axis_idx, x=None, y=None):
        """
        Args:
        'idx':  profile index (int)
        'axis_idx': axis index (int [self._N<=2] or tuple [self._N>2])
        'x' and 'y': data arrays
        """
        indexes = super()._set_axis_indexes(axis_idx)
        if x==None and y==None:
            self._axis[indexes].plot(self._radius[idx], self._relax[idx])
        else:
            self._axis[indexes].plot(x, y)

    @do_all_objects
    def plot_all(self, x, y, *args, **kwargs):
        self.plot(kwargs['i_profile'], (kwargs['i'], kwargs['j']))

    def intersect_and_plot(self, idx, axis_idx, intersect_value=1.):
        indexes = super()._set_axis_indexes(axis_idx)
        degree = 5
        self._axis[indexes].plot(self._radius[idx], self._relax[idx], '.') 
        _intersect_x = intersect(x=self._radius[idx], y=self._relax[idx], deg=degree)
        _polyfit = np.polyfit(self._radius[idx], self._relax[idx], deg=degree)
        _func = np.poly1d(_polyfit)
        _x_new = np.linspace(self._radius[idx][0], self._relax[idx][-1], 10000)
        _y_new = _func(_x_new)
        self._axis[indexes].plot(_x_new, _y_new, markersize=1, label='polynomial fit') #draw pseudo-line
        self._axis[indexes].plot([min(self._radius[idx]), max(self._radius[idx])], 
                                 [intersect_value, intersect_value], 
                                 color='r', linestyle='--', 
                                 label=r't$_{relax}(R)$/t$_{circ}$(r$_{200}$) = '+str(intersect_value))
        _half_bin_radius = (self._radius[idx][1]-self._radius[idx][0])/2
        _half_bin_relax = (self._relax[idx][1]-self._relax[idx][0])/2
        _text_idx = len(self._relax[idx])-int(len(self._relax[idx])/4)
        if _intersect_x[1]==True: #if there was an intersection
            self._axis[indexes].plot(_intersect_x[0], intersect_value, 'ko')
            self._axis[indexes].annotate('Intersection at x='+str(np.round_(_intersect_x[0],3)), xy=(0.015,0.75), xycoords='axes fraction')
        self._axis[indexes].set_xlim([self._radius[idx][0] -_half_bin_radius, 
                                      self._radius[idx][-1] + _half_bin_radius])
        self._axis[indexes].set_ylim([self._relax[idx][0] + _half_bin_relax, 
                                      self._relax[idx][-1] - _half_bin_relax])
        self._axis[indexes].legend()

    @do_all_objects
    def intersect_and_plot_all(self, function='nfw', *args, **kwargs):
        self.intersect_and_plot(kwargs['i_profile'], (kwargs['i'], kwargs['j']))



import numpy as np
from pynbody.analysis.theoretical_profiles import NFWprofile

def nfw_fit(prof):
    """
    NFW fit. Returns the characteristic density and the scale radius.
    """
    _fit = NFWprofile.fit(np.array(prof['rbins']),
                          np.array(prof['density']))
    print("FIT: ", _fit)
    return _fit[0][0], _fit[0][1]



from distutils.core import setup

setup(
    name = 'dmprofile',
    packages = ['dmprofile'],
    version = 'v0.1.0',
    description = 'pynbody wrapper to facilitate standard analysis of dark matter halos',
    author = 'b-fontana',
    author_email = 'bruno.afonso.alves@ist.utl.pt',
    url = 'https://github.com/b-fontana/DarkMatter',
    download_url = 'https://github.com/b-fontana/DarkMatter/archive/v0.1.0.tar.gz',
    keywords = ['dark matter', 'halos', 'pynbody'],
    classifiers = [],
)
